/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_cellprofiler_javabridge_CPython */

#ifndef _Included_org_cellprofiler_javabridge_CPython
#define _Included_org_cellprofiler_javabridge_CPython
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_cellprofiler_javabridge_CPython
 * Method:    exec
 * Signature: (Ljava/lang/String;Ljava/util/Map;Ljava/util/Map;)V
 */
JNIEXPORT void JNICALL Java_org_cellprofiler_javabridge_CPython_exec
  (JNIEnv *, jobject, jstring, jobject, jobject);

#ifdef __cplusplus
}
#endif
#endif
